use axum::{
    async_trait,
    extract::{FromRef, FromRequest, State},
    http::{Request, StatusCode},
    Json,
};
use ed25519_dalek::{Signature, VerifyingKey};
use hyper::Body;
use sqlx::PgPool;
use twilight_interactions::command::{CommandInputData, CommandModel, CreateCommand};
use twilight_model::{
    application::interaction::{Interaction, InteractionData, InteractionType},
    http::interaction::{InteractionResponse, InteractionResponseType},
};

use crate::discord::commands::{get_fact, set_fact, GetFactCommand, SetFactCommand, RollFactCommand, roll_fact};

pub struct ExtractInteraction(pub Interaction);

#[async_trait]
impl<S> FromRequest<S, Body> for ExtractInteraction
where
    S: Send + Sync,
    VerifyingKey: FromRef<S>,
{
    type Rejection = (StatusCode, String);

    async fn from_request(request: Request<Body>, state: &S) -> Result<Self, Self::Rejection> {
        let (parts, body) = request.into_parts();
        let body_bytes = hyper::body::to_bytes(body)
            .await
            .map_err(|err| (StatusCode::INTERNAL_SERVER_ERROR, err.to_string()))?;

        let Ok(interaction): Result<Interaction, _> = serde_json::from_slice(&body_bytes) else {
            return Err((StatusCode::BAD_REQUEST, "request contained invalid json".to_string()))
        };
        let Some(sig) = parts.headers.get("x-signature-ed25519") else {
            return Err((StatusCode::BAD_REQUEST, "request did not include signature header".to_string()))
        };
        let Ok(sig) = hex::decode(sig) else {
            return Err((StatusCode::BAD_REQUEST, "requrest signature is invalid hex".to_string()))
        };
        let Ok(sig) = Signature::from_slice(&sig) else {
            return Err((StatusCode::BAD_REQUEST, "request signature is malformed".to_string()))
        };
        let Some(signed_buf) = parts.headers.get("x-signature-timestamp") else {
            return Err((StatusCode::BAD_REQUEST, "requrest did not include signature timestamp header".to_string()))
        };
        let mut signed_buf = signed_buf.as_bytes().to_owned();
        signed_buf.extend(body_bytes);

        let public_key = VerifyingKey::from_ref(state);
        let Ok(()) = public_key.verify_strict(&signed_buf, &sig) else {
            return Err((StatusCode::UNAUTHORIZED, "interaction failed signature verification".to_string()))
        };

        Ok(Self(interaction))
    }
}

type InteractionResult = Result<(StatusCode, Json<InteractionResponse>), (StatusCode, String)>;

pub async fn post_interaction(
    State(pg_pool): State<PgPool>,
    ExtractInteraction(interaction): ExtractInteraction,
) -> InteractionResult {
    match interaction.kind {
        InteractionType::Ping => {
            let pong = InteractionResponse {
                kind: InteractionResponseType::Pong,
                data: None,
            };

            Ok((StatusCode::OK, Json(pong)))
        }
        InteractionType::ApplicationCommand => {
            let author_id = interaction.author_id();
            let Some(InteractionData::ApplicationCommand(data)) = interaction.data.clone() else {
                return not_found();
            };
            let command_input_data = CommandInputData::from(*data.clone());
            tracing::debug!(command_name = ?data.name, "started processing command");
            let result = match &*data.name {
                SetFactCommand::NAME => {
                    let Ok(command_data) = SetFactCommand::from_interaction(command_input_data) else {
                        return Err((StatusCode::BAD_REQUEST, format!("invalid {0} command.", SetFactCommand::NAME)));
                    };
                    let Some(author_id) = author_id else {
                        return Err((StatusCode::BAD_REQUEST, format!("{0} requires a user.", SetFactCommand::NAME)));
                    };
                    match set_fact(
                        interaction.id,
                        interaction.channel_id,
                        author_id,
                        command_data,
                        &pg_pool,
                    )
                    .await
                    {
                        Ok(response) => Ok((StatusCode::OK, Json(response))),
                        Err(err) => Err(err),
                    }
                }
                GetFactCommand::NAME => {
                    let Ok(command_data) = GetFactCommand::from_interaction(command_input_data) else {
                        return Err((StatusCode::BAD_REQUEST, format!("invalid {0} command.", GetFactCommand::NAME)));
                    };
                    let Some(author_id) = author_id else {
                        return Err((StatusCode::BAD_REQUEST, format!("{0} requires a user.", GetFactCommand::NAME)));
                    };

                    match get_fact(interaction.channel_id, author_id, command_data, &pg_pool).await
                    {
                        Ok(response) => Ok((StatusCode::OK, Json(response))),
                        Err(err) => Err(err),
                    }
                }
                RollFactCommand::NAME => {
                    let Ok(command_data) = RollFactCommand::from_interaction(command_input_data) else {
                        return Err((StatusCode::BAD_REQUEST, format!("invalid {0} command.", SetFactCommand::NAME)));
                    };
                    let Some(author_id) = author_id else {
                        return Err((StatusCode::BAD_REQUEST, format!("{0} requires a user.", SetFactCommand::NAME)));
                    };
                    match roll_fact(
                        interaction.id,
                        interaction.channel_id,
                        author_id,
                        command_data,
                        &pg_pool,
                    )
                    .await
                    {
                        Ok(response) => Ok((StatusCode::OK, Json(response))),
                        Err(err) => Err(err),
                    }
                }
                _ => not_found()
            };

            tracing::debug!(command_name = ?data.name, "finished processing command");
            result
        }
        _ => not_found(),
    }
}

fn not_found() -> InteractionResult {
    Err((
        StatusCode::NOT_FOUND,
        "requested interaction not found".to_string(),
    ))
}
