use std::{net::SocketAddr, process};

use anyhow::bail;
use axum::{routing::post, Router};

#[cfg(feature = "ngrok")]
use ngrok::{tunnel::UrlTunnel, prelude::TunnelBuilder};
use sqlx::postgres::PgPoolOptions;
use tokio::signal;

use tower_http::trace::TraceLayer;
use tracing_subscriber::{
    filter::LevelFilter, layer::SubscriberExt, util::SubscriberInitExt, Layer,
};

use god_replacement_product::{
    discord::{commands::register_commands, interactions::post_interaction},
    AppState, Config,
};

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let config = Config::configure()?;
    let tracer = opentelemetry_jaeger::new_agent_pipeline()
        .with_service_name("god_replacement_product")
        .install_simple()?;
    let telemetry = tracing_opentelemetry::layer().with_tracer(tracer);
    let (loki_layer, loki_task) = tracing_loki::builder()
        .label("host", "mine")?
        .extra_field("pid", format!("{}", process::id()))?
        .build_url(url::Url::parse("http://127.0.0.1:3100").unwrap())?;

    tracing_subscriber::registry()
        .with(telemetry.with_filter(LevelFilter::DEBUG))
        .with(
            tracing_subscriber::fmt::layer()
                .and_then(loki_layer)
                .with_filter(
                    tracing_subscriber::EnvFilter::try_from_default_env().unwrap_or_else(|_| {
                        "god_replacement_product=debug,tower_http=debug".into()
                    }),
                ),
        )
        .init();

    let pg_pool = PgPoolOptions::new()
        .max_connections(5)
        .connect(&config.database_url)
        .await?;

    sqlx::migrate!().run(&pg_pool).await?;

    register_commands(config.discord_client_id.to_owned(), config.authorization()).await?;
    let state = AppState {
        config: config.clone(),
        pg_pool,
    };

    let app = Router::new()
        .route("/api/discord/interactions/", post(post_interaction))
        .with_state(state)
        .layer(TraceLayer::new_for_http());
    tokio::spawn(loki_task);

    if config.ngrok_use {
        #[cfg(feature = "ngrok")]
        {
            let mut tunnel_builder = ngrok::Session::builder()
                .authtoken_from_env()
                .connect()
                .await?
                .http_endpoint();
            if let Some(domain) = config.ngrok_domain {
                tunnel_builder = tunnel_builder.domain(domain);
            }

            let tunnel = tunnel_builder.listen().await?;
            tracing::debug!("Ngrok tunnel at {:?}", tunnel.url());
            axum::Server::builder(tunnel)
                .serve(app.into_make_service())
                .with_graceful_shutdown(shutdown_signal())
                .await?;
        }
        #[cfg(not(feature = "ngrok"))]
        {
            bail!("Ngrok requested, but feature ngrok wasn't enabled.")
        }
    } else if let Some(port) = config.listen_port {
        let addr = SocketAddr::from(([127, 0, 0, 1], port));
        tracing::debug!("listening on {}", addr);

        axum::Server::bind(&addr)
            .serve(app.into_make_service())
            .with_graceful_shutdown(shutdown_signal())
            .await?;
    } else {
        bail!("Failed to construct a builder");
    };

    tracing::info!("server down");

    opentelemetry::global::shutdown_tracer_provider();

    Ok(())
}

async fn shutdown_signal() {
    let ctrl_c = async {
        signal::ctrl_c()
            .await
            .expect("failed to install Ctrl+C handler");
    };

    #[cfg(unix)]
    let terminate = async {
        signal::unix::signal(signal::unix::SignalKind::terminate())
            .expect("failed to install signal handler")
            .recv()
            .await;
    };

    #[cfg(not(unix))]
    let terminate = std::future::pending::<()>();

    tokio::select! {
        _ = ctrl_c => {},
        _ = terminate => {},
    }

    tracing::info!("signal received, starting graceful shutdown");
}
